@@ -23,7 +23,7 @@ or with a python-like shorthand
 
 - catherinedevlin.blogspot.com  May 31, 2006
 """
-import sys, os, re, sqlpython, cx_Oracle, pyparsing, re, completion, datetime, pickle, binascii
+import sys, os, re, sqlpython, cx_Oracle, pyparsing, re, completion, datetime, pickle, binascii, subprocess
 from cmd2 import Cmd, make_option, options, Statekeeper, Cmd2TestCase
 from output_templates import output_templates
 from plothandler import Plot
@@ -364,11 +364,13 @@ class sqlpyPlus(sqlpython.sqlpython):
     def __init__(self):
         sqlpython.sqlpython.__init__(self)
         self.binds = CaselessDict()
-        self.settable = 'autobind commit_on_exit echo maxfetch maxtselctrows timeout'.split()
+        self.settable += 'autobind commit_on_exit maxfetch maxtselctrows timeout heading'.split()
+        self.settable.sort()
         # settables must be lowercase
         self.stdoutBeforeSpool = sys.stdout
         self.spoolFile = None
         self.autobind = False
+        self.heading = True
     #def default(self, arg):
     #    sqlpython.sqlpython.default(self, arg)
 
@@ -442,7 +444,7 @@ class sqlpyPlus(sqlpython.sqlpython):
             plot.draw()
             return ''
         else:
-            result = sqlpython.pmatrix(self.rows, self.curs.description, self.maxfetch)
+            result = sqlpython.pmatrix(self.rows, self.curs.description, self.maxfetch, heading=self.heading)
         return result
         
     legalOracle = re.compile('[a-zA-Z_$#]')
@@ -535,32 +537,133 @@ class sqlpyPlus(sqlpython.sqlpython):
             print '\nSelected Max Num rows (%d)' % self.rc
         
     def do_cat(self, arg):
+        '''Shortcut for SELECT * FROM'''
         return self.do_select(self.parsed('SELECT * FROM %s;' % arg, 
                                           terminator = arg.parsed.terminator or ';', 
                                           suffix = arg.parsed.suffix))
-        
-    @options([make_option('-f', '--full', action='store_true', help='get dependent objects as well')])
+
+    def _pull(self, arg, opts, vc=None):
+        """Displays source code."""
+        if opts.dump:
+            statekeeper = Statekeeper(self, ('stdout',))                        
+        try:
+            for (owner, object_type, object_name) in self.resolve_many(arg, opts):  
+                if object_type in self.supported_ddl_types:
+                    object_type = {'DATABASE LINK': 'DB_LINK', 'JAVA CLASS': 'JAVA_SOURCE'
+                                   }.get(object_type) or object_type
+                    object_type = object_type.replace(' ', '_')
+                    if opts.dump:
+                        try:
+                            os.makedirs(os.path.join(owner.lower(), object_type.lower()))
+                        except OSError:
+                            pass
+                        filename = os.path.join(owner.lower(), object_type.lower(), '%s.sql' % object_name.lower())
+                        self.stdout = open(filename, 'w')
+                        if vc:
+                            subprocess.call(vc + [filename])
+                    try:
+                        if object_type in ['CONTEXT', 'DIRECTORY', 'JOB']:
+                            ddlargs = [object_type, object_name]
+                        else:
+                            ddlargs = [object_type, object_name, owner]
+                        self.stdout.write(str(self.curs.callfunc('DBMS_METADATA.GET_DDL', cx_Oracle.CLOB, ddlargs)))
+                    except cx_Oracle.DatabaseError:
+                        if object_type == 'JOB':
+                            print '%s: DBMS_METADATA.GET_DDL does not support JOBs (MetaLink DocID 567504.1)' % object_name
+                            continue
+                        raise
+                    if opts.full:
+                        for dependent_type in ('OBJECT_GRANT', 'CONSTRAINT', 'TRIGGER'):        
+                            try:
+                                self.stdout.write(str(self.curs.callfunc('DBMS_METADATA.GET_DEPENDENT_DDL', cx_Oracle.CLOB,
+                                                                         [dependent_type, object_name, owner])))
+                            except cx_Oracle.DatabaseError:
+                                pass
+                    if opts.dump:
+                        self.stdout.close()
+        except:
+            if opts.dump:
+                statekeeper.restore()
+            raise
+        if opts.dump:
+            statekeeper.restore()    
+
+    def do_show(self, arg):
+        '''
+        show                  - display value of all sqlpython parameters
+        show (parameter name) - display value of a sqlpython parameter
+        show parameter (parameter name) - display value of an ORACLE parameter
+        '''
+        if arg.startswith('param'):
+            try:
+                paramname = arg.split()[1].lower()
+            except IndexError:
+                paramname = ''
+            self.onecmd("""SELECT name, 
+                           CASE type WHEN 1 THEN 'BOOLEAN'
+                                     WHEN 2 THEN 'STRING'
+                                     WHEN 3 THEN 'INTEGER'
+                                     WHEN 4 THEN 'PARAMETER FILE'
+                                     WHEN 5 THEN 'RESERVED'
+                                     WHEN 6 THEN 'BIG INTEGER' END type, 
+                           value FROM v$parameter WHERE name LIKE '%%%s%%';""" % paramname)
+        else:
+            return Cmd.do_show(self, arg)
+            
+    @options([make_option('-d', '--dump', action='store_true', help='dump results to files'),
+              make_option('-f', '--full', action='store_true', help='get dependent objects as well'),
+              make_option('-a', '--all', action='store_true', help="all schemas' objects"),
+              make_option('-x', '--exact', action='store_true', help="match object name exactly")])
     def do_pull(self, arg, opts):
         """Displays source code."""
+        self._pull(arg, opts)
+            
+    supported_ddl_types = 'CLUSTER, CONTEXT, DATABASE LINK, DIRECTORY, FUNCTION, INDEX, JOB, LIBRARY, MATERIALIZED VIEW, PACKAGE, PACKAGE BODY, OPERATOR, PACKAGE, PROCEDURE, SEQUENCE, SYNONYM, TABLE, TRIGGER, VIEW, TYPE, TYPE BODY, XML SCHEMA'
+    do_pull.__doc__ += '\n\nSupported DDL types: ' + supported_ddl_types
+    supported_ddl_types = supported_ddl_types.split(', ')    
 
-        target = arg.upper()
-        object_type, owner, object_name = self.resolve(target)
-        if not object_type:
-            return
-        self.stdout.write("%s %s.%s\n" % (object_type, owner, object_name))
-        self.stdout.write(str(self.curs.callfunc('DBMS_METADATA.GET_DDL', cx_Oracle.CLOB,
-                                                 [object_type, object_name, owner])))
-        if opts.full:
-            for dependent_type in ('OBJECT_GRANT', 'CONSTRAINT', 'TRIGGER'):        
-                try:
-                    self.stdout.write(str(self.curs.callfunc('DBMS_METADATA.GET_DEPENDENT_DDL', cx_Oracle.CLOB,
-                                                             [dependent_type, object_name, owner])))
-                except cx_Oracle.DatabaseError:
-                    pass
+    def _vc(self, arg, opts, program):
+        subprocess.call([program, 'init'])
+        opts.dump = True
+        self._pull(arg, opts, vc=[program, 'add'])
+        subprocess.call([program, 'commit', '-m', '"%s"' % opts.message or 'committed from sqlpython'])        
+    
+    @options([
+              make_option('-f', '--full', action='store_true', help='get dependent objects as well'),
+              make_option('-a', '--all', action='store_true', help="all schemas' objects"),
+              make_option('-x', '--exact', action='store_true', help="match object name exactly"),
+              make_option('-m', '--message', action='store', type='string', dest='message', help="message to save to hg log during commit")])
+    def do_hg(self, arg, opts):
+        '''hg (opts) (objects):
+        Stores DDL on disk and puts files under Mercurial version control.
+        Args specify which objects to store, same format as `ls`.'''
+        self._vc(arg, opts, 'hg')        
 
+    @options([
+              make_option('-f', '--full', action='store_true', help='get dependent objects as well'),
+              make_option('-a', '--all', action='store_true', help="all schemas' objects"),
+              make_option('-x', '--exact', action='store_true', help="match object name exactly"),
+              make_option('-m', '--message', action='store', type='string', dest='message', help="message to save to hg log during commit")])
+    def do_bzr(self, arg, opts):
+        '''bzr (opts) (objects):
+        Stores DDL on disk and puts files under Bazaar version control.
+        Args specify which objects to store, same format as `ls`.'''
+        self._vc(arg, opts, 'bzr')        
+
+    @options([
+              make_option('-f', '--full', action='store_true', help='get dependent objects as well'),
+              make_option('-a', '--all', action='store_true', help="all schemas' objects"),
+              make_option('-x', '--exact', action='store_true', help="match object name exactly"),
+              make_option('-m', '--message', action='store', type='string', dest='message', help="message to save to hg log during commit")])
+    def do_git(self, arg, opts):
+        '''git (opts) (objects):
+        Stores DDL on disk and puts files under git version control.
+        Args specify which objects to store, same format as `ls`.'''
+        self._vc(arg, opts, 'git')        
+        
     all_users_option = make_option('-a', action='store_const', dest="scope",
-                                         default={'col':'', 'view':'user', 'schemas':'user'}, 
-                                         const={'col':', owner', 'view':'all', 'schemas':'all'},
+                                         default={'col':'', 'view':'user', 'schemas':'user', 'firstcol': ''}, 
+                                         const={'col':', owner', 'view':'all', 'schemas':'all', 'firstcol': 'owner, '},
                                          help='Describe all objects (not just my own)')                
     @options([all_users_option,
               make_option('-c', '--col', action='store_true', help='find column'),
@@ -571,11 +674,11 @@ class sqlpyPlus(sqlpython.sqlpython):
         capArg = arg.upper()
         
         if opts.col:
-            sql = "SELECT table_name, column_name %s FROM %s_tab_columns where column_name like '%%%s%%';" \
-                % (opts.scope['col'], opts.scope['view'], capArg)
+            sql = "SELECT table_name, column_name %s FROM %s_tab_columns where column_name like '%%%s%%' ORDER BY %s table_name, column_name;" \
+                % (opts.scope['col'], opts.scope['view'], capArg, opts.scope['firstcol'])
         elif opts.table:
-            sql = "SELECT table_name %s from %s_tables where table_name like '%%%s%%';" \
-                % (opts.scope['col'], opts.scope['view'], capArg)
+            sql = "SELECT table_name %s from %s_tables where table_name like '%%%s%%' ORDER BY %s table_name;" \
+                % (opts.scope['col'], opts.scope['view'], capArg, opts.scope['firstcol'])
         else:
             sql = "SELECT * from %s_source where UPPER(text) like '%%%s%%';" % (opts.scope['view'], capArg)
         self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
@@ -794,6 +897,9 @@ class sqlpyPlus(sqlpython.sqlpython):
         self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
 
     def do_head(self, arg):
+        '''Shortcut for SELECT * FROM <arg>;10
+        The terminator (\\t, \\g, \\x, etc.) and number of rows can
+        be changed as for any other SELECT statement.'''
         sql = self.parsed('SELECT * FROM %s;' % arg, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix)
         sql.parsed['suffix'] = sql.parsed.suffix or '10'
         self.do_select(self.parsed(sql))
@@ -868,14 +974,16 @@ class sqlpyPlus(sqlpython.sqlpython):
 
     def do_declare(self, arg):
         self.anon_plsql('declare ' + arg)
-
-    @options([make_option('-l', '--long', action='store_true', help='long descriptions'),
-              make_option('-a', '--all', action='store_true', help="all schemas' objects")])        
-    def do_ls(self, arg, opts):
-        where = ''
+        
+    def _ls_statement(self, arg, opts):
         if arg:
-            where = """\nWHERE object_type || '/' || object_name
-                  LIKE '%%%s%%'""" % (arg.upper().replace('*','%'))
+            target = arg.upper()
+            if hasattr(opts, 'exact') and opts.exact:
+                where = """\nWHERE object_name = '%s'
+                             OR object_type || '/' || object_name = '%s'""" % \
+                            (target, target)
+            else:
+                where = "\nWHERE object_type || '/' || object_name LIKE '%%%s%%'" % (arg.upper().replace('*','%'))
         else:
             where = ''
         if opts.all:
@@ -884,13 +992,41 @@ class sqlpyPlus(sqlpython.sqlpython):
         else:
             whose = 'user'
             objname = 'object_name'            
-        if opts.long:
-            extraInfo = ', status, last_ddl_time AS modified'
+        if hasattr(opts, 'long') and opts.long:
+            moreColumns = ', status, last_ddl_time'
         else:
-            extraInfo = ''
-        statement = '''SELECT object_type || '/' || %s AS name %s 
-                  FROM   %s_objects %s
-                  ORDER BY object_type, object_name;''' % (objname, extraInfo, whose, where)
+            moreColumns = ''
+            
+        # 'Normal' sort order is DATE DESC (maybe), object type ASC, object name ASC
+        sortdirection = ('DESC' if hasattr(opts, 'reverse') and opts.reverse else 'ASC')
+        orderby = 'object_type %s, object_name %s' % (sortdirection, sortdirection)
+        if hasattr(opts, 'timesort') and opts.timesort:
+            orderby = 'last_ddl_time %s, %s' % (('ASC' if hasattr(opts, 'reverse') and opts.reverse else 'DESC'), orderby)
+        return {'objname': objname, 'moreColumns': moreColumns,
+                'whose': whose, 'where': where, 'orderby': orderby}        
+        
+    def resolve_many(self, arg, opts):
+        opts.long = False
+        clauses = self._ls_statement(arg, opts)
+        if opts.all:
+            clauses['owner'] = 'owner'
+        else:
+            clauses['owner'] = 'user'
+        statement = '''SELECT %(owner)s, object_type, object_name 
+                  FROM   %(whose)s_objects %(where)s
+                  ORDER BY object_type, object_name''' % clauses
+        self.curs.execute(statement)
+        return self.curs.fetchall()
+    
+    @options([make_option('-l', '--long', action='store_true', help='long descriptions'),
+              make_option('-a', '--all', action='store_true', help="all schemas' objects"),
+              make_option('-t', '--timesort', action='store_true', help="Sort by last_ddl_time"),              
+              make_option('-r', '--reverse', action='store_true', help="Reverse order while sorting"),              
+              make_option('-x', '--exact', action='store_true', default=False, help="match name exactly")])        
+    def do_ls(self, arg, opts):
+        statement = '''SELECT object_type || '/' || %(objname)s AS name %(moreColumns)s 
+                  FROM   %(whose)s_objects %(where)s
+                  ORDER BY %(orderby)s;''' % self._ls_statement(arg, opts)
         self.do_select(self.parsed(statement, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
         
     @options([make_option('-i', '--ignore-case', dest='ignorecase', action='store_true', help='Case-insensitive search')])        
